use super::service::Buffer;
use std::{fmt, marker::PhantomData};
use tower_layer::Layer;
use tower_service::Service;

/// Adds an mpsc buffer in front of an inner service.
///
/// The default Tokio executor is used to run the given service,
/// which means that this layer can only be used on the Tokio runtime.
///
/// See the module documentation for more details.
pub struct BufferLayer<Request, E2 = crate::BoxError> {
    bound: usize,
    _p: PhantomData<fn(Request, E2)>,
}

impl<Request, E2> BufferLayer<Request, E2> {
    /// Creates a new `BufferLayer` with the provided `bound`.
    ///
    /// `bound` gives the maximal number of requests that can be queued for the service before
    /// backpressure is applied to callers.
    ///
    /// # A note on choosing a `bound`
    ///
    /// When `Buffer`'s implementation of `poll_ready` returns `Poll::Ready`, it reserves a
    /// slot in the channel for the forthcoming `call()`. However, if this call doesn't arrive,
    /// this reserved slot may be held up for a long time. As a result, it's advisable to set
    /// `bound` to be at least the maximum number of concurrent requests the `Buffer` will see.
    /// If you do not, all the slots in the buffer may be held up by futures that have just called
    /// `poll_ready` but will not issue a `call`, which prevents other senders from issuing new
    /// requests.
    pub fn new(bound: usize) -> Self {
        BufferLayer {
            bound,
            _p: PhantomData,
        }
    }
}

impl<S, Request, E2> Layer<S> for BufferLayer<Request, E2>
where
    S: Service<Request> + Send + 'static,
    S::Future: Send,
    S::Error: Clone + Into<E2> + Send + Sync + std::fmt::Display,
    Request: Send + 'static,
    E2: Send + 'static,
    crate::buffer::error::Closed: Into<E2>,
    Request: Send + 'static,
{
    type Service = Buffer<S, Request, E2>;

    fn layer(&self, service: S) -> Self::Service {
        Buffer::new2(service, self.bound)
    }
}

impl<Request, E2> fmt::Debug for BufferLayer<Request, E2> {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        f.debug_struct("BufferLayer")
            .field("bound", &self.bound)
            .finish()
    }
}
